/**********************************************************************/
/*! \class RtMidi
    \brief An abstract base class for realtime MIDI input/output.

    This class implements some common functionality for the realtime
    MIDI input/output subclasses RtMidiIn and RtMidiOut.

    RtMidi WWW site: http://music.mcgill.ca/~gary/rtmidi/

    RtMidi: realtime MIDI i/o C++ classes
    Copyright (c) 2003-2010 Gary P. Scavone

		Modified by S Jagannathan for use in din
		July 2010.
*/
/**********************************************************************/

// RtMidi: Version 1.0.11

#ifndef RTMIDI_H
#define RTMIDI_H

#include "RtError.h"
#include <string>

// ALSA header file.
#include <alsa/asoundlib.h>

// A structure to hold variables related to the ALSA API
// implementation.
struct AlsaMidiData {
  snd_seq_t *seq;
  int vport;
  snd_seq_port_subscribe_t *subscription;
  snd_midi_event_t *coder;
  int bufferSize;
  unsigned char buffer[256];
  pthread_t thread;
  unsigned long long lastTime;
  int queue_id; // an input queue is needed to get timestamped events
};

class RtMidi {
 
 public:

  virtual void openPort( unsigned int portNumber = 0, const std::string portName = std::string( "RtMidi" ) ) = 0;
  virtual unsigned int getPortCount() = 0;
  virtual std::string getPortName( unsigned int portNumber = 0 ) = 0;
  virtual void closePort( void ) = 0;

  RtMidi();
  virtual ~RtMidi() {};

  void error( RtError::Type type );
  AlsaMidiData apiData_;
  bool connected_;
  std::string errorString_;

};

#include <vector>
#include <queue>

class RtMidiIn : public RtMidi
{
 public:

  typedef void (*RtMidiCallback)( double timeStamp, std::vector<unsigned char> *message, void *userData);
  RtMidiIn( const std::string clientName = std::string( "RtMidi Input Client") );
  ~RtMidiIn();
  void openPort( unsigned int portNumber = 0, const std::string Portname = std::string( "RtMidi Input" ) );
  void setCallback( RtMidiCallback callback, void *userData = 0 );
  void cancelCallback();
  void closePort( void );
  unsigned int getPortCount();
  std::string getPortName( unsigned int portNumber = 0 );
  void setQueueSizeLimit( unsigned int queueSize );
  void ignoreTypes( bool midiSysex = true, bool midiTime = true, bool midiSense = true );
  double getMessage( std::vector<unsigned char> *message );
  
  struct MidiMessage { 
    std::vector<unsigned char> bytes; 
    double timeStamp;
    MidiMessage() :bytes(3), timeStamp(0.0) {}
  };

  struct RtMidiInData {
    std::queue<MidiMessage> queue;
    MidiMessage message;
    unsigned int queueLimit;
    unsigned char ignoreFlags;
    bool doInput;
    bool firstMessage;
    AlsaMidiData *apiData;
    bool usingCallback;
    void *userCallback;
    void *userData;
    bool continueSysex;
    RtMidiInData () : queueLimit(1024), ignoreFlags(7), doInput(false), firstMessage(true), 
    apiData(0), usingCallback(false), userCallback(0), userData(0), continueSysex(false) {
		}
  };

 private:

  void initialize( const std::string& clientName );
  RtMidiInData inputData_;

};


#endif
