/*

 * din
 * copyright(c) 2006-2010 Sampath Jagannathan
 * distributed under GNU 2.0 license. Please see file COPYING for more information.
 * 

*/
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <tcl.h>
#include "box.h"
#include "utils.h"
#include "font.h"
#include "console.h"
#include "input.h"
#include "viewwin.h"
#include "curve.h"
#include "multi_curve.h"
#include "solver.h"
#include "chrono.h"
#include "ui.h"
#include "curve_editor.h"
#include "font_editor.h"
#include "din.h"
#include "curve_library.h"
#include "console_iterator.h"
#include "delay.h"
#include "random.h"
#include "globals.h"
#include "command.h"
#include "scalelist.h"
#include "fft.h"
#include "sine_mixer.h"
#include "checkdotdin.h"
#include "osc.h"
#include "bot.h"
#include "morse_code.h"
#include "tcl_interp.h"
#include <vector>
#include <iostream>
#include <fstream>
#include <list>
#include <map>
#include <algorithm>

using namespace std;

// ~/.din contains defaults, user savings and preferences
string dotdin; 
checkdotdin cdd;

// tcl interpreter. all din commands are available in the interpreter.
tcl_interp interpreter;

// custom vector font
font fnt ("jag.fnt");

// console
console cons;

// audio output - using JACK
audio_out aout;

// display - using OpenGL
viewport view;

// initial globals
load_globals lg;

// list of scales
scalelist scalelst;

// keyboard and mouse state
keyboard keybd;
int mousex, mousey;
int lmb, rmb, mmb;

// L and R delays
delay left_delay (1000, 0.75, 0.3, "feedback-l.crv", "volume-l.crv"), right_delay (1000, 0.75, 0.3, "feedback-r.crv", "volume-r.crv");
curve_editor delayed ("delay.ed");

void setup_delay_editor () {
	delayed.add (&left_delay.fbk_crv, &left_delay.fbk_lis);
	delayed.add (&left_delay.vol_crv, &left_delay.vol_lis);
	delayed.add (&right_delay.fbk_crv, &right_delay.fbk_lis);
	delayed.add (&right_delay.vol_crv, &right_delay.vol_lis);
}

// midi bpm got from external midi clock
double MIDI_BPM = 0;

// din clock
chrono clk;

fft fft0; // fast fourier transform (via fftwf)
sine_mixer sinemixer; // sine mixer for waveform edit

cmdlist cmdlst; // list of din commands

// din. see din.cc/.h
din din0 (cmdlst);

// global commands, available from all editors
//
// format: long name, short name, help or help_file_name
//

show_help help ("help", "?", "help <command_name>", &cmdlst);

// for L and R delays
set_delay sd (&left_delay, &right_delay, "set_delay", "sd", "set_delay.hlp");
get_delay gd (&left_delay, &right_delay, "get_delay", "gd", "get_delay.hlp");

// for adding/listing/removing scales
add_scale as ("add_scale", "as", "add_scale.hlp");
ls_scale ls ("ls_scale", "ls", "list_scale.hlp");
remove_scale rs ("remove_scale", "rs", "remove_scale.hlp");

// set and get din var
setv sv (&din0, "setv", "sv", "setv/sv <var_name> <value>");
getv gv (&din0, "getv", "gv", "getv/gv <var_name>");

// set/get bpm
set_bpm sb ("set_bpm", "sb", "set_bpm/sb <gate/gl/gr/am/fm/os> <value>");
get_bpm gb ("get_bpm", "gb", "get_bpm/gb <gate-l/gate-r/am/fm/os> <value>");

// set/get bpm style
set_style ss ("set_style", "ss", "set_style/ss <gate/gl/gr/am/fm/os> <loop/pong>");
get_style gs ("get_style", "gs", "get_style/gs <gl/gr/am/fm/os>");

// on/off gaters
gaters gt (&din0, "gaters", "g", "gaters/g <on/off>");

// set key/tonic
key ky (&din0, "key", "sa", "key.hlp");

// load scale
load_scale los (&din0, "load_scale", "los", "load_scale.hlp");

// load/save resonators that make din sound
load_resonators lre (&din0, "load_resonators", "lre", "load_resonators/lre < <file_name> > ; eg, lre resonators");
save_resonators sre (&din0, "save_resonators", "sre", "save_resonators/sre [file_name] [delete_disabled_resonators] ; eg, sre resonators 1 ; deletes disabled resonators and saves remaining resonators to file ~/.din/resonators");
delete_disabled_resonators ddr (&din0, "delete_disabled_resonators", "ddr", "delete_disabled_resonators/ddr ; deletes disabled resonators. no undo");

// display notation on the keyboard
notation no (&din0, "notation", "no", "notation.hlp");

// music utils
note_distance nd ("note_distance", "nd", "note_distance <note1> <note2>; nd C G");
chord ch ("chord", "ch", "chord.hlp");

// font cmds
set_font_size sfs ("set_font_size", "sfs", "set_font_size <xsize> <ysize> <charspc>; eg., sfs 4 3 4;");
get_font_size gfs ("get_font_size", "gfs", "get_font_size");
set_kern sk ("set_kern", "sk", "set_kern <chara> <charb> <amt>; eg., sk C a -1");
get_kern gk ("get_kern", "gk", "get_kern <chara> <charb>; eg., gk i j");

// curve cmds
curve_name cn ("curve_name", "cn", "curve_name.hlp");
curve_value cv ("curve_value", "cv", "curve_value.hlp");
curve__library cl ("curve_library", "cl", "curve_library.hlp");
set_curve_editor sced ("set_curve_editor", "sed", "set_curve_editor.hlp");
get_curve_editor gced ("get_curve_editor", "ged", "get_curve_editor.hlp");

// morse code
morse_code mc ("morse_code", "mc", "morse_code.hlp");

// unix like echo
echo ech ("echo", "say", "echo/say <string>. prints string on the console");

// OSC server/client commands
lo_address client  = 0;
oserver srv ("oserver", "re", "oserver.hlp");
oconnect co (client, "connect", "co", "connect <ip> <port>; opens an osc connection on ip at port. eg. co 83.22.199.1 3333");
osend snd (client, "send", "snd", "send <path> <format> <data>");

// IRC bot
bot bott ("bot", ":", "bot.hlp");

quit qu ("quit", "exit", "quit/exit - quits din");

midi mid ("midi", "midi", "midi support in din");

/*struct work_t {
	int start, end, fpb;
	float buf [audio_out::SAMPLES_PER_BUFFER];
};

const int MAX_CORES = 8;
work_t work [MAX_CORES];
pthread_t tid [MAX_CORES];

void* dowork (void* arg) {
	
	work_t* w = (work_t*) arg;
	
  memset (w->buf, 0, audio_out::FBUFFER_SIZE); 
	
	int start = w->start, end = w->end, fpb = w->fpb;
	float* buf = w->buf;
	for (int i = start; i < end; ++i) {
		float* buf1 = buf;
		din0.players[i] (buf1, fpb, din0.ams, din0.fms);
	}
	return 0;
}*/

void addcmds () {
	cmdlst.add (&help);
	cmdlst.add (&sd);
	cmdlst.add (&gd);
	cmdlst.add (&as);
	cmdlst.add (&ls);
	cmdlst.add (&rs);
	cmdlst.add (&nd);
	cmdlst.add (&ch);
	cmdlst.add (&sk);
	cmdlst.add (&gk);
	cmdlst.add (&sfs);
	cmdlst.add (&gfs);
	cmdlst.add (&co);
	cmdlst.add (&snd);
	cmdlst.add (&srv);
	cmdlst.add (&ech);
	cmdlst.add (&cn);
	cmdlst.add (&cv);
	cmdlst.add (&cl);
	cmdlst.add (&mc);
	cmdlst.add (&bott);
	cmdlst.add (&sv);
	cmdlst.add (&gv);
	cmdlst.add (&sb);
	cmdlst.add (&gb);
	cmdlst.add (&ss);
	cmdlst.add (&gs);
	cmdlst.add (&gt);
	cmdlst.add (&ky);
	cmdlst.add (&los);
	cmdlst.add (&lre);
	cmdlst.add (&sre);
	cmdlst.add (&ddr);
	cmdlst.add (&no);
	cmdlst.add (&qu);
	cmdlst.add (&sced);
	cmdlst.add (&gced);
	cmdlst.add (&mid);
}

uis_t uis;

void set_window (int w, int h) {

	extern string APP_NAME;
	extern int SCREEN_DEPTH;

	view = viewport (APP_NAME, w, h, SCREEN_DEPTH);
	w = view.width; h = view.height;

	fft0.lev.chkpos ();
	sinemixer.sine_levels.chkpos ();

	--w; --h;
	cons.set_window (box<int>(0, -h, w, 0));

	din0.win (0, 0, w, h);

	din0.prev_mousex = -1;
	din0.prev_mousey = -1;

	din0.enter ();
		
}

void load_window () {

	extern int SCREEN_WIDTH, SCREEN_HEIGHT;
	int width = SCREEN_WIDTH, height = SCREEN_HEIGHT;
	string ignore;

	extern string dotdin;
	ifstream file ((dotdin + "window").c_str(), ios::in);
	if (file) {
		file >> ignore >> width;
		file >> ignore >> height;
	}
	set_window (width, height);
}

void save_window () {
	const char* fname = "window";
	extern string dotdin;
	ofstream file ((dotdin + fname).c_str(), ios::out);
	extern viewport view;
	if (file) {
		file << "width " << view.width << endl;
		file << "height " << view.height << endl;
	}	
}

bool read_input () {
	
    SDL_Event event;
    while (SDL_PollEvent(&event)) {
      switch(event.type) {
        case SDL_VIDEORESIZE:
          try {
            set_window (event.resize.w, event.resize.h);
          } catch (const badviewport& bv) {
            cout << "sdl: viewport init failed" << endl;
            return false;
          }
          break;
      }
    }
		
    keybd.read ();
    
    int buttons = SDL_GetMouseState (&mousex, &mousey);
		lmb = buttons & SDL_BUTTON_LMASK;
		rmb = buttons & SDL_BUTTON_RMASK;
		mmb = buttons & SDL_BUTTON_MMASK; 
		
    return true;

}

int quit_now = 0;

int main (int argc, char** argv) {
	
	system (". ~/.din/m00"); // xset m 0 0 for pure mouse data, without any fudge applied by X.
  setSeed (clock()); 
  addcmds ();
  setup_delay_editor ();
  cons.home ();
  load_window ();
	uis.set_current (&din0, 0);

	string load_settings_tcl ("source " + dotdin + "settings.tcl");
	interpreter (load_settings_tcl);
	
  aout.start ();
	
  while (1) {
    if (!read_input() || !uis.handle_input () || quit_now) break;
    uis.draw ();
		osc ();
		mid.handle_input ();
    keybd.save ();
    Tcl_DoOneEvent (TCL_DONT_WAIT);
  }
  
  aout.stop ();
	save_window ();
	system ("xset m $ACCEL $THRESHOLD"); // restore mouse settings
  return 0;
}

int jack_audio_wanted (jack_nframes_t nsamples, void *arg) {
	
	audio_out* aout = (audio_out *) arg;
	
	void* midi_buf = jack_port_get_buffer (aout->midi_in, nsamples);
	jack_nframes_t nevents = jack_midi_get_event_count (midi_buf);
	static int kkk = 0; 
	if (nevents > 1) {
		cout << ++kkk << " nevents = " << nevents << eol;
	}
	
	sample_t* lbuf = (sample_t *) jack_port_get_buffer (aout->out_left, nsamples); // left buf
	sample_t* rbuf = (sample_t *) jack_port_get_buffer (aout->out_right, nsamples); // right buf
	
	// zero out buffers
	memset (lbuf, 0, aout->sample_buffer_size);
	memset (rbuf, 0, aout->sample_buffer_size);
	
	// find tone/volume from mouse position, take contrib from each resonator, do fm/am
	if (din0.find_tone_and_volume ()) {
		
		din0.fill (nsamples);
		for (int i = 0; i < din0.num_resonators; ++i) {
			float* rl = lbuf, *rr = rbuf;
			din0.players[i] (rl, rr, nsamples, din0.result, din0.ams, din0.fms);
		}
				
	}
	
	// apply gaters
	//
	
	if (din0.gatl.enabled) { // left gater
		float* gl = lbuf;
		din0.gatl (gl, din0.result, nsamples, 1);
	}

	if (din0.gatr.enabled) { // right gater
		float* gr = rbuf;
		din0.gatr (gr, din0.result, nsamples, 1);
	}	
	
	// apply octave shift
	//
	if (din0.is_octave_shift) {
		din0.octave_shift (din0.result, nsamples);
		static int last = nsamples - 1;
		if (din0.osh.dir) din0.set_tonic (din0.osh.tonic * din0.result[last]); else din0.set_tonic (din0.osh.tonic / din0.result[last]);
	}
	
	for (int i = 0, j = din0.num_drones; i < j; ++i) {
		float* dl = lbuf, *dr = rbuf;
		play& pi = din0.drones[i].player;
		pi (dl, dr, nsamples, din0.result);
	}
	
	// apply delay
	//
	
	if (din0.dinfo.delay) {
		float* outl = lbuf; left_delay (outl, nsamples, 1);
		float* outr = rbuf; right_delay (outr, nsamples, 1);
	}
		
		
	// apply compression
	//
	if (din0.dinfo.compress) {
		float* outl = lbuf;
		float* outr = rbuf;
		din0.coml.apply (outl, nsamples, 1);
		din0.comr.apply (outr, nsamples, 1);
	}		
	
	// write audio to oscilloscope
	if (!din0.osc.paused && din0.osc.visible) {
		float* outl = lbuf;
		float* outr = rbuf;
		din0.osc.add_samples (outl, outr, nsamples);
	}
	
	clk += nsamples;	
	
	return 0;
	
}

extern int SAMPLE_RATE;
extern double SECS_PER_BUFFER;

int sample_rate_changed (jack_nframes_t nframes, void *arg) {
	audio_out* aout = (audio_out*) arg;
	aout->sample_rate = jack_get_sample_rate (aout->client);
	SAMPLE_RATE = (int) aout->sample_rate;
	SECS_PER_BUFFER = jack_get_buffer_size (aout->client) * 1./SAMPLE_RATE;
	din0.update_range_notes ();
	din0.update_drone_tone ();
	return 0;
}


int buffer_size_changed (jack_nframes_t nframes, void *arg) {
	audio_out* aout = (audio_out*) arg;
	aout->sample_buffer_size = sizeof(sample_t) * jack_get_buffer_size (aout->client);
	SECS_PER_BUFFER = nframes * 1./SAMPLE_RATE;
	return 0;
}
